'' =================================================================================================
''
''   File....... jm_pwm595x2.spin
''   Purpose.... PWM modulation for 595 x 2 circuits
''   Author..... Jon "JonnyMac" McPhalen 
''               Copyright (c) 2011-13 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 20 NOV 2013
''
'' =================================================================================================


con

  CHANNELS = 16
  LAST_CH  = CHANNELS - 1
  

var

  long  cog                                                     ' cog running driver

  long  eout                                                    ' output enable pin (if used)
  
  byte  brightness[CHANNELS]
 
    
pub start(dpin, cpin, lpin, epin) 

'' Starts 595x2 PWM driver
'' -- dpin connects to 74x595.SER (1st device)
'' -- cpin connects to 74x595.SCK
'' -- lpin connects to 74x595.RCK
'' -- epin connects to 74x595.G (optional)

  stop

  ' setup cog

  odatmask   := 1 << dpin                                       ' create pin masks   
  oclkmask   := 1 << cpin
  olatchmask := 1 << lpin

  eout := epin
  if (eout => 0)                                                ' using output enable?   
    outa[eout] := 1                                             '  disable at startup    
    dira[eout] := 1
  
  hubaddr := @brightness                                        ' set address of brightness levels
  
  cog := cognew(@entry, 0) + 1                                  ' start the cog

  return cog


pub stop

'' Stops cog (if running)

  if (cog)
    cogstop(cog - 1)
    cog := 0
    if (eout => 0)
      dira[eout] := 0
  
  set_all(0)


pub enable

'' Enable x595 outputs

  if (eout => 0)
    outa[eout] := 0


pub disable

'' Disable x595 outputs

  if (eout => 0)
    outa[eout] := 1


pub set(ch, level)

'' Sets channel to specified level

  if ((ch => 0) & (ch =< LAST_CH))
    brightness[ch] := 0 #> level <# 255

  return brightness[ch]
    

pub set_all(level)

'' Set all channels to same level

  level := 0 #> level <# 255                                    ' limit to byte value
  bytefill(@brightness[0], level, CHANNELS)

  return level

  
pub inc(ch)

'' Increment channel brightness

  if ((ch => 0) & (ch =< LAST_CH))
    if (brightness[ch] < 255)
      ++brightness[ch]

  return brightness[ch]


pub dec(ch)

'' Decrement channel brightness

  if ((ch => 0) & (ch =< LAST_CH))
    if (brightness[ch] > 0)
      --brightness[ch]

  return brightness[ch]
      

pub high(ch)

'' Sets channel on

  if ((ch => 0) & (ch =< LAST_CH))
    brightness[ch] := 255


pub low(ch)

'' Sets channel off

  if ((ch => 0) & (ch =< LAST_CH))
    brightness[ch] := 0


pub toggle(ch)

'' Toggles channel
'' -- threshold for analog value is 50%

  if ((ch => 0) & (ch =< LAST_CH))
    if (brightness[ch] < 128)
      high(ch)
    else
      low(ch)    

  
pub ez_log(level)

'' Creates log-like curve for LEDs
'' -- improves apparent linearity of brightness

  level := 0 #> level <# 255                                    ' limit to byte value
  
  return (level * level) >> 8                                   ' level := (level^2) / 256


pub read(ch)

'' Returns channel brightness

  if ((ch => 0) & (ch =< LAST_CH))
    return brightness[ch]
  else
    return -1


pub address

'' Returns hub address of brightness array

  return @brightness


dat

                        org     0

entry                   mov     outa, #0                        ' all off
                        or      dira, odatmask                  ' set pins to output
                        or      dira, oclkmask
                        or      dira, olatchmask

                        or      dira, testpin

                        mov     cycle, #0
                        
pwmmain                 xor     outa, testpin                   ' toggle test bit

                        mov     hub, hubaddr                    ' point to brightness array
                        mov     chcount, #16                    ' read 16 channels
                        mov     outbits, #0                     ' reset working outputs
                        mov     chmask, #1                      ' starting mask 

:loop                   rdbyte  t1, hub                 wz      ' get channel value, save 0
        if_z            jmp     #:next                          ' skip if zero
                        cmp     cycle, t1               wc, wz  ' check level
        if_be           or      outbits, chmask                 ' on if not done

:next                   add     hub, #1                         ' next value
                        shl     chmask, #1
                        djnz    chcount, #:loop

                        
shiftout16              mov     bitcount, #16                   ' shift out 16 bits
                        shl     outbits, #(32-16)               ' prep for msb first

:loop                   shl     outbits, #1             wc      ' get msb
                        muxc    outa, odatmask                  ' output the bit
                        nop                                     ' let bit settle
                        or      outa, oclkmask                  ' clock the bit
                        nop
                        andn    outa, oclkmask
                        djnz    bitcount, #:loop
                        or      outa, olatchmask                ' latch the outputs
                        nop
                        andn    outa, olatchmask

                        add     cycle, #1                       ' next cycle value
                        and     cycle, #$FF                     ' keep 0 to 255

                        jmp     #pwmmain 

' --------------------------------------------------------------------------------------------------

testpin                 long    1 << 16


odatmask                long    0-0                             ' pin mask for dat
oclkmask                long    0-0                             ' pin mask for clk
olatchmask              long    0-0                             ' pin mask for latch
hubaddr                 long    0-0                             ' hub address of brightness array

hub                     res     1
cycle                   res     1
chcount                 res     1
outbits                 res     1                               ' bits to shift out
chmask                  res     1
bitcount                res     1                               ' bit count

t1                      res     1                               ' work vars
t2                      res     1

                        fit     496                                    

                        
dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}  